{% set jquery_ui_includes = ['base', 'multiselect'] %}
{% extends 'templating:admin/layout.html' %}
{% block title %}{{ _('Manage People') }}{% endblock%}
{% block body %}
    <h2>{{ _('Manage People') }}</h2>
    {{ people_grid.html()|content }}
{% endblock body%}
